import random

import pytest

from debputy.packager_provided_files import detect_all_packager_provided_files
from debputy.path_matcher import ExactFileSystemPath
from debputy.substitution import NULL_SUBSTITUTION
from .util import FakeDirEntry, faked_binary_package, binary_package_table


@pytest.mark.parametrize("package_name,basename,install_target", [
    ('foo', 'foo.tmpfiles', './usr/lib/tmpfiles.d/foo.conf'),
    # main package match
    ('foo', 'tmpfiles', './usr/lib/tmpfiles.d/foo.conf'),
    # arch match
    ('foo', 'foo.tmpfiles.amd64', './usr/lib/tmpfiles.d/foo.conf'),
    # Matches with periods in both package name and in the file type
    ('foo.bar', 'foo.bar.ppp.ip-up',  './etc/ppp/ip-up.d/foo.bar'),
    ('foo.bar', 'ppp.ip-up',  './etc/ppp/ip-up.d/foo.bar'),
    ('foo.bar', 'foo.bar.ppp.ip-up.amd64',  './etc/ppp/ip-up.d/foo.bar'),

    # Priority
    ('foo.bar', 'foo.bar.udev',  './lib/udev/rules.d/60-foo.bar'),
    ('foo.bar', 'udev', './lib/udev/rules.d/60-foo.bar'),
    ('foo.bar', 'foo.bar.udev.amd64', './lib/udev/rules.d/60-foo.bar'),

    # Name
    ('foo.bar', 'foo.bar.special.name.udev', './lib/udev/rules.d/60-special.name'),
    ('foo.bar', 'foo.bar.special.name.udev.amd64', './lib/udev/rules.d/60-special.name'),
])
def test_packager_provided_files(debputy_plugin_feature_set, package_name, basename, install_target):
    dir_entry = FakeDirEntry(basename, directory='debian')
    binary_packages = binary_package_table(
        faked_binary_package(package_name)
    )
    ppf = detect_all_packager_provided_files(debputy_plugin_feature_set.packager_provided_files,
                                             [dir_entry],
                                             binary_packages,
                                             lambda n: NULL_SUBSTITUTION,
                                             )
    assert package_name in ppf
    all_matched = ppf[package_name]
    assert len(all_matched) == 1
    matched = all_matched[0]
    assert matched.source_path == dir_entry.path
    assert f'debian/{basename}' in matched.definition_source
    match_rule = matched.match_rule
    assert isinstance(match_rule, ExactFileSystemPath)
    assert match_rule.path == install_target


@pytest.mark.parametrize("package_name,expected_basename,non_matched", [
    ('foo', 'foo.tmpfiles', ['tmpfiles']),
    ('foo', 'foo.tmpfiles.amd64', ['foo.tmpfiles', 'tmpfiles', 'foo.tmpfiles.i386']),
    ('foo', 'foo.tmpfiles', ['tmpfiles', 'foo.tmpfiles.i386']),
])
def test_packager_provided_files_priority(debputy_plugin_feature_set, package_name, expected_basename, non_matched):
    assert len(non_matched) > 0

    dir_entry = FakeDirEntry(expected_basename, directory='debian')
    binary_packages = binary_package_table(
        faked_binary_package(package_name)
    )
    all_entries_base = [
        FakeDirEntry(x, directory='debian')
        for x in non_matched
    ]
    for order in (0, len(all_entries_base), None):
        all_entries = all_entries_base.copy()
        print(f"Order: {order}")
        if order is not None:
            all_entries.insert(order, dir_entry)
        else:
            all_entries.append(dir_entry)
            # Ensure there are no order dependencies in the test by randomizing it.
            random.shuffle(all_entries)
        ppf = detect_all_packager_provided_files(debputy_plugin_feature_set.packager_provided_files,
                                                 all_entries,
                                                 binary_packages,
                                                 lambda n: NULL_SUBSTITUTION,
                                                 )
        assert package_name in ppf
        all_matched = ppf[package_name]
        assert len(all_matched) == 1
        matched = all_matched[0]
        assert matched.source_path == dir_entry.path
        assert f'debian/{expected_basename}' in matched.definition_source
        match_rule = matched.match_rule
        assert isinstance(match_rule, ExactFileSystemPath)
